FUNCTION linear_regression_zphot_tarrio, features_training, features_test, z_spec, Nneigh, z_phot_err, flag_extrapolation

  ; ************************************************
  ; Linear regression to calculate z_phot for all the elements in the test set
  ; ************************************************
  Ntrain = n_elements(z_spec)
  size_test = size(features_test)
  Ntest = size_test[1]

  neighbours = lonarr(Ntest,Nneigh)
  flag_extrapolation = fltarr(Ntest,1)
  z_phot = fltarr(Ntest,1)
  z_phot_err = fltarr(Ntest,1)

  for i=0,Ntest-1 do begin
    
    ; Search for the 100 closest training points
    dist_to_training = total((features_test[i,*] ## transpose(fltarr(1,Ntrain)+1.) - features_training)^2.,2)
    ; FAST SORT: we reduce the array size before to the sort
    index_short_dist = where(dist_to_training lt 0.02,ss)
    if ss lt Nneigh then begin
      index_short_dist = where(dist_to_training lt 0.05,rr)
      if rr lt Nneigh then begin
        index_short_dist = where(dist_to_training lt 0.1,qq)
        if qq lt Nneigh then begin
          index_short_dist = where(dist_to_training lt 0.5,nn)
          if nn lt Nneigh then begin
            index_short_dist = where(dist_to_training lt 1.,mm)
            if mm lt Nneigh then begin
              index_short_dist = where(dist_to_training lt 10.,oo)
              if oo lt Nneigh then begin
                index_short_dist = where(dist_to_training lt 210.,pp)
                if pp lt Nneigh then index_short_dist = lindgen(n_elements(dist_to_training))
              endif
            endif
          endif
        endif
      endif
    endif
    aux = dist_to_training[index_short_dist]
    index_sorted = index_short_dist[sort(aux)]
    neighbours[i,*] = index_sorted[0:Nneigh-1]
    
    ; Linear regression to estimate z_phot
    xx = features_training[neighbours[i,*],*]
    yy = z_spec[neighbours[i,*]]
    H = [[xx],[fltarr(Nneigh)+1.]]
    res = (invert((H)##transpose(H))##H)##yy
    z_phot[i] = total(res*[[features_test[i,*]],[1]])
    z_phot_neigh = transpose(res)##H
    z_phot_err[i] = sqrt(total((yy-z_phot_neigh)^2.)/Nneigh)
    if total(features_test[i,*] lt min(xx,dimension=1) or features_test[i,*] gt max(xx,dimension=1)) gt 0 then flag_extrapolation[i] = 1
    
    ; Refinement (see Beck et al 2016, section 2.1, penultimate paragraph):
    neigh_index = where( (yy-z_phot_neigh) le 3.*z_phot_err[i], ng) ; Neighbours not outliers
    if ng lt Nneigh and ng gt 0 then begin
      xx = xx[neigh_index,*]
      yy = yy[0,neigh_index]
      H = [[xx],[fltarr(ng)+1.]]
      res = (invert((H)##transpose(H))##H)##yy
      z_phot[i] = total(res*[[features_test[i,*]],[1]])
      z_phot_neigh = transpose(res)##H
      z_phot_err[i] = sqrt(total((yy-z_phot_neigh)^2.)/ng)
      if total(features_test[i,*] lt min(xx,dimension=1) or features_test[i,*] gt max(xx,dimension=1)) gt 0 then flag_extrapolation[i] = 1
    endif
    
  endfor

  return,z_phot

END